from callback.Callback import Callback
from Task import Task
from Timer import Timer
from matrix_wrapper import *
from callback import callback_helper
from callback.callbacks import TaskCallback


class TaskCallback(Callback):
    ACTION_NAME = "task"
    WRONG_ID_MESSAGE = "there was a wrong id"

    async def run(self, args: dict):
        actions: list[str] = await callback_helper.get_multi_params(args, Callback.NO_PARAM, TaskCallback.ACTION_NAME)
        if actions is None: # actions is at least an empty list; actually not needed, but leaved for simplicity
            return
        elif not actions:
            await send_html("<code>{}</code> needs <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code> or <code>{}</code>".format(
                TaskCallback.ACTION_NAME, Callback.CREATE_PARAM, Callback.GET_PARAM, Callback.DELETE_PARAM, Callback.TIMER_PARAM, 
                Callback.SUBTASK_PARAM, Callback.TICK_PARAM, Callback.UNTICK_PARAM
            ))
            return

        action = actions.pop(0)

        if action == Callback.CREATE_PARAM:
            await self.handle_create(actions, args)

        elif action == Callback.GET_PARAM:
            await self.handle_get(actions, args)

        elif action == Callback.DELETE_PARAM:
            await self.handle_delete(actions, args)

        elif action == Callback.TIMER_PARAM:
            await self.handle_timer(actions, args)

        elif action == Callback.SUBTASK_PARAM:
            await self.handle_subtask(actions, args)

        elif action == Callback.TICK_PARAM:
            await self.handle_tick(actions, args)

        elif action == Callback.UNTICK_PARAM:
            await self.handle_untick(actions, args)

        elif action == Callback.DUMP_PARAM:
            await self.handle_dump(actions, args)

        else:
            await send_message("unknown param {}".format(action))

        return await super().run(args)

    async def handle_create(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskCallback.ACTION_NAME, Callback.CREATE_PARAM)
        task_names = await callback_helper.get_multi_params(args, Callback.NAME_ARG, pre_command)
        if task_names is None:
            return

        message = "created tasks"
        for task_name in task_names:
            task = Task(task_name)
            message += " {}".format(task.get_id())
        
        await send_message(message)

    async def handle_get(self, no_params: list[str], args: dict): # subtasks
        if args.get(Callback.SUBTASKS_ARG) is not None:
            await self.handle_get_with_subtasks(no_params, args)
        else:
            await self.handle_get_without_subtasks(no_params, args)        

    async def handle_get_without_subtasks(self, no_params: list[str], args: dict):
        task_ids = args.get(Callback.ID_ARG)
        tasks = self._get_tasks_from_ids(task_ids)

        if tasks is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return

        message = """<h2>List Tasks</h2>
<table>
    <tr>
        <th>Task ID</th>
        <th>Task name</th>
        <th>Timer count</th>
        <th>Subtask count</th>
        <th>State</th>
    </tr>
        """
            
        with Task.tasks_lock:
            for task in tasks:
                message += "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>"\
                .format(task.get_id(), task.get_name(), len(task.get_timers()), len(task.get_subtasks()),
                Task.task_state_to_html(task.get_state()))

        message += "</table>"

        await send_html(message)

    async def handle_get_with_subtasks(self, no_params: list[str], args: dict):
        task_ids = args.get(Callback.ID_ARG)
        if task_ids is None or not task_ids:
            await send_html("<code>{} {}</code> with param <code>{}</code> needs arg <code>{}</code>".format(TaskCallback.ACTION_NAME, Callback.GET_PARAM, Callback.ID_ARG))
            return
        
        tasks: list[Task] = self._get_tasks_from_ids(task_ids)
        if tasks is None:
            await send_html(TaskCallback.WRONG_ID_MESSAGE)
            return

        message = """<h2>List Tasks with Subtasks</h2>
<ul>
        """

        for task in tasks:
            message += self.get_subtasks_str(task)

        message += "</ul>"

        await send_html(message)

    async def handle_delete(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskCallback.ACTION_NAME, Callback.DELETE_PARAM)
        task_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command)
        
        tasks: list[Task] = []
        for task_id in task_ids:
            task = Task.get_from_id(task_id)
            if task is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                return
            tasks.append(task)

        message = "deleted tasks"
        for task in tasks:
            message += " {}".format(task.get_id())
            task.remove()

        await send_message(message)

    async def handle_timer(self, no_params: list[str], args: dict):
        no_wrong_param_message = "<code>{} {}</code> needs <code>{}</code>, <code>{}</code> or <code>{}</code>".format(TaskCallback.ACTION_NAME, Callback.TIMER_PARAM, Callback.CREATE_PARAM, Callback.LINK_PARAM, Callback.GET_PARAM)
        if not no_params:
            await send_html(no_wrong_param_message)
        timer_action = no_params.pop(0)

        if timer_action == Callback.CREATE_PARAM:
            await self.handle_timer_create(no_params, args)
        elif timer_action == Callback.LINK_PARAM:
            await self.handle_timer_link(no_params, args)
        elif timer_action == Callback.GET_PARAM:
            await self.handle_timer_get(no_params, args)
        else:
            await send_html(no_wrong_param_message)

    async def handle_timer_create(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.TIMER_PARAM, Callback.CREATE_PARAM)
        timer_text = await callback_helper.get_single_param(args, Callback.TEXT_ARG, pre_command, required=False)
        timer_duration = await callback_helper.get_single_param(args, Callback.DURATION_ARG, pre_command, float)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)

        if timer_duration is None or task_id is None:
            return

        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return
        
        timer = task.add_timer(timer_duration, timer_text)
        await send_message("created timer {} and added it to task {}".format(timer.get_id(), task.get_id()))

    async def handle_timer_link(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.TIMER_PARAM, Callback.LINK_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)
        timer_ids = await callback_helper.get_multi_params(args, Callback.TIMER_ID_ARG, pre_command)

        if task_id is None or timer_ids is None:
            return
        
        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return
        
        timers: list[Timer] = []
        for timer_id in timer_ids:
            timer = Timer.get_from_id(timer_id)
            if timer is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                return

            timers.append(timer)

        message = "linked timers"
        for timer in timers:
            task.link_timer(timer)
            message += " {}".format(timer.get_id())

        message += " to task {}".format(task.get_id())

        await send_message(message)

    async def handle_timer_get(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.TIMER_PARAM, Callback.GET_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)

        if task_id is None:
            return

        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return

        message = """<h2>List Timers of Task "{}"</h2>
<table>
    <tr>
        <th>Timer ID</th>
        <th>remaining</th>
    </tr>
        """.format(task.get_name())

        for timer in task.get_timers():
            message += "<tr><td>{}</td><td>{}</td></tr>".format(timer.get_id(), str(timer.remaining()))

        message += "<table>"
        await send_html(message)

    async def handle_subtask(self, no_params: list[str], args: dict):
        no_wrong_param_message = "<code>{} {}</code> needs <code>{}</code>, <code>{}</code>, <code>{}</code> or <code>{}</code>".format(
            TaskCallback.ACTION_NAME, Callback.SUBTASK_PARAM, Callback.CREATE_PARAM, Callback.LINK_PARAM, Callback.UNLINK_PARAM, Callback.GET_PARAM)
        if not no_params:
            await send_html(no_wrong_param_message)
        subtask_action = no_params.pop(0)

        if subtask_action == Callback.CREATE_PARAM:
            await self.handle_subtask_create(no_params, args)
        elif subtask_action == Callback.LINK_PARAM:
            await self.handle_subtask_link(no_params, args)
        elif subtask_action == Callback.GET_PARAM:
            await self.handle_subtask_get(no_params, args)
        elif subtask_action == Callback.UNLINK_PARAM:
            await self.handle_subtask_unlink(no_params, args)
        else:
            await send_html(no_wrong_param_message)

    async def handle_subtask_create(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.SUBTASK_PARAM, Callback.CREATE_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)
        subtask_names = await callback_helper.get_multi_params(args, Callback.NAME_ARG, pre_command)

        if task_id is None or subtask_names is None:
            return

        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return

        message = "created subtasks"
        for subtask_name in subtask_names:
            subtask = task.add_subtask(subtask_name)
            message += " {}".format(subtask.get_id())

        message += " and added them to task {}".format(task.get_id())
        
        await send_message(message)

    async def handle_subtask_link(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.SUBTASK_PARAM, Callback.LINK_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)
        subtask_ids = await callback_helper.get_multi_params(args, Callback.SUBTASK_ID_ARG, pre_command)

        if task_id is None or subtask_ids is None:
            return
        
        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return
        
        subtasks: list[Task] = []
        for subtask_id in subtask_ids:
            subtask = Task.get_from_id(subtask_id)
            if subtask is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                
            subtasks.append(subtask)

        message = "linked subtasks"
        for subtask in subtasks:
            task.link_subtask(subtask)
            message += " {}".format(subtask.get_id())

        message += " to task {}".format(task.get_id())

        await send_message(message)

    async def handle_subtask_get(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.SUBTASK_PARAM, Callback.GET_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)

        if task_id is None:
            return

        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return

        message = """<h2>List Subtasks of Task "{}"
<table>
    <tr>
        <th>Subtask ID</th>
        <th>Name</th>
        <th>Has Subtasks</th>
        <th>Has Timers</th>
        <th>State</th>
    </tr>
        """.format(task.get_name())

        for subtask in task.get_subtasks():
            message += "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>".format(
                subtask.get_id(), str(subtask.get_name()), bool(task.get_subtasks()), bool(task.get_timers()), Task.task_state_to_html(task.get_state()))

        message += "<table>"
        await send_html(message)

    async def handle_subtask_unlink(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskCallback.ACTION_NAME, Callback.SUBTASK_PARAM, Callback.UNLINK_PARAM)
        task_id = await callback_helper.get_single_param(args, Callback.TASK_ID_ARG, pre_command)
        subtask_ids = await callback_helper.get_multi_params(args, Callback.SUBTASK_ID_ARG, pre_command)

        if task_id is None or subtask_ids is None:
            return
        
        task = Task.get_from_id(task_id)
        if task is None:
            await send_message(TaskCallback.WRONG_ID_MESSAGE)
            return
        
        subtasks: list[Task] = []
        for subtask_id in subtask_ids:
            subtask = Task.get_from_id(subtask_id)
            if subtask is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                
            subtasks.append(subtask)

        message = "unlinked subtasks"
        for subtask in subtasks:
            task.unlink_subtask(subtask)
            message += " {}".format(subtask.get_id())

        message += " from task {}. remove them manually if you want to".format(task.get_id())

        await send_message(message)

    async def handle_tick(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskCallback.ACTION_NAME, Callback.TICK_PARAM)
        task_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command)

        if task_ids is None:
            return

        message = "ticked tasks"
        for task_id in task_ids:
            task = Task.get_from_id(task_id)
            if task is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                return

            task.tick()
            message += " {}".format(task.get_id())

        await send_message(message)

    async def handle_untick(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskCallback.ACTION_NAME, Callback.UNTICK_PARAM)
        task_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command)

        if task_ids is None:
            return

        message = "unticked tasks"
        for task_id in task_ids:
            task = Task.get_from_id(task_id)
            if task is None:
                await send_message(TaskCallback.WRONG_ID_MESSAGE)
                return

            task.untick()
            message += " {}".format(task.get_id())

        await send_message(message)

    async def handle_dump(self, no_params: list[str], args: dict):
        Task.dump_all()

    def get_subtasks_str(self, task: Task):
        subtasks = task.get_subtasks()
        message = "<li>{} {}: {}".format(Task.task_state_to_html(task.get_state()), task.get_name(), task.get_id())
        for subtask in subtasks:
            message += "<ul>{}</ul>".format(self.get_subtasks_str(subtask))

        message += "</li>"
        return message

    def _get_tasks_from_ids(self, task_ids: list[str]) -> list[Task]:
        tasks = []
        if task_ids is not None:
            for task_id in task_ids:
                task = Task.get_from_id(task_id)
                if task is None:
                    return None

                tasks.append(task)
                
        else:
            with Task.tasks_lock:
                tasks = Task.tasks

        return tasks
        
