from TaskGroup import TaskGroup
from Task import Task
from callback.callbacks import TaskGroupCallback
from callback.callbacks.TaskCallback import TaskCallback
from callback.Callback import Callback
from callback import callback_helper
from matrix_wrapper import *

class TaskGroupCallback(Callback):
    ACTION_NAME = "tgroup"
    WRONG_ID_MESSAGE = "there was a wrong id"

    PATH: list[TaskGroup] = []
    CURRENT_TASKGROUP: TaskGroup = None

    async def run(self, args: dict):
        actions: list[str] = await callback_helper.get_multi_params(args, Callback.NO_PARAM, TaskGroupCallback.ACTION_NAME)
        if actions is None: # actions is at least an empty list; actually not needed, but leaved for simplicity
            return
        elif not actions:
            await send_html("<code>{}</code> needs <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code>, <code>{}</code> or <code>{}</code>".format(
                TaskGroupCallback.ACTION_NAME, Callback.CREATE_PARAM, Callback.GET_PARAM, Callback.DELETE_PARAM, Callback.TIMER_PARAM, 
                Callback.SUBTASK_PARAM, Callback.TICK_PARAM, Callback.UNTICK_PARAM
            ))
            return

        ## todo
        if callback_helper._interactive:
            args.update({Callback.ID_ARG: [TaskGroupCallback.CURRENT_TASKGROUP.get_id(),]})

        action = actions.pop(0)

        if action == Callback.CREATE_PARAM:
            await self.handle_create(actions, args)

        elif action == Callback.GET_PARAM:
            await self.handle_get(actions, args)

        elif action == Callback.DELETE_PARAM:
            await self.handle_delete(actions, args)

        elif action == Callback.TASK_PARAM:
            await self.handle_task(actions, args)

        elif action == Callback.SUBTASKGROUP_PARAM:
            await self.handle_subtaskgroup(actions, args)

        elif action == Callback.DUMP_PARAM:
            await self.handle_dump(actions, args)

        # elif action == Callback.INTERACTIVE_PARAM:
        #     await self.handle_interactive_set(actions, args)

        else:
            await send_message("unknown param {}".format(action))

        return await super().run(args)

    async def handle_create(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskGroupCallback.ACTION_NAME, Callback.CREATE_PARAM)
        group_names = await callback_helper.get_multi_params(args, Callback.NAME_ARG, pre_command)
        if group_names is None:
            return

        message = "created taskgroups"
        for group_name in group_names:
            taskGroup = TaskGroup(group_name)
            message += " {}".format(taskGroup.get_id())

        await send_message(message)

    async def handle_get(self, no_params: list[str], args: dict):
        if args.get(Callback.SUBGROUPS_ARG) is not None:
            await self.handle_get_with_subgroups(no_params, args)
        else:
            await self.handle_get_without_subgroups(no_params, args)

    async def handle_get_with_subgroups(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskGroupCallback.ACTION_NAME, Callback.GET_PARAM)
        group_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command)
        subtask_arg = await callback_helper.get_multi_params(args, Callback.SUBTASKS_ARG, pre_command, False)
        with_tasks = subtask_arg is not None
        if group_ids is None:
            return

        groups = self._get_taskgroup_from_id(group_ids)
        if groups is None:
            await send_html(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        message = """<h2>List TaskGroups with Subgroups</h2>
<ul>
        """

        for taskGroup in groups:
            message += self.get_subtasks_group_str(taskGroup, with_tasks)

        message += "</ul>"
        await send_html(message)

    async def handle_get_without_subgroups(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskGroupCallback.ACTION_NAME, Callback.GET_PARAM)
        taskgroup_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command, False)
        taskgroups = self._get_taskgroup_from_id(taskgroup_ids)

        message = """<h2>List TaskGroups</h2>
<table>
    <tr>
        <th>TaskGroup ID</th>
        <th>TaskGroup Name</th>
        <th>SubTaskGroup count</th>
        <th>Task count</th>
    </tr>
        """

        for taskgroup in taskgroups:
            message += "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>"\
                .format(taskgroup.get_id(), taskgroup.get_name(), len(taskgroup.get_subTaskGroups()), len(taskgroup.get_tasks()))

        message += "</table>"
        await send_html(message)

    async def handle_delete(self, no_params: list[str], args: dict):
        pre_command = "{} {}".format(TaskGroupCallback.ACTION_NAME, Callback.DELETE_PARAM)
        taskgroup_ids = await callback_helper.get_multi_params(args, Callback.ID_ARG, pre_command, False)

        message = "deleted TaskGroups"
        taskgroups: list[TaskGroup] = []
        for taskgroup_id in taskgroup_ids:
            taskgroup = TaskGroup.get_from_id(taskgroup_id)
            if taskgroup is None:
                await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
                return
            taskgroups.append(taskgroup)

        message = "deleted TaskGroups"
        for taskgroup in taskgroups:
            message += " {}".format(taskgroup.get_id())
            taskgroup.remove()

        await send_message(message)

    async def handle_task(self, no_params: list[str], args: dict):
        if not no_params:
            await send_html("<code>{} {}</code> needs <code>{}</code>, <code>{}</code>, <code>{}</code> or <code>{}</code>".format(
                TaskGroupCallback.ACTION_NAME, Callback.TASK_PARAM, Callback.CREATE_PARAM, Callback.GET_PARAM, Callback.LINK_PARAM, Callback.UNLINK_PARAM
            ))
            return

        action = no_params.pop(0)

        if action == Callback.CREATE_PARAM:
            await self.handle_task_create(no_params, args)
        elif action == Callback.GET_PARAM:
            await self.handle_task_get(no_params, args)
        elif action == Callback.LINK_PARAM:
            await self.handle_task_link(no_params, args)
        elif action == Callback.UNLINK_PARAM:
            await self.handle_task_unlink(no_params, args)
        else:
            await send_message("unknown param {}".format(action))


    async def handle_task_create(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.TASK_PARAM, Callback.CREATE_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        task_names = await callback_helper.get_multi_params(args, Callback.NAME_ARG, pre_command)

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        message = "created tasks"
        for task_name in task_names:
            task = group.add_task(task_name)
            message += " {}".format(task.get_id())
        
        message += " and added them to TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_task_get(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.TASK_PARAM, Callback.GET_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        if group_id is None:
            return

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return
            

        message="""<h2>Tasks of TaskGroup "{}"</h2>
<ul>
        """.format(group.get_name())

        for task in group.get_tasks():
            message += TaskCallback().get_subtasks_str(task)

        message += "</ul>"
        await send_html(message)

    async def handle_task_link(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.TASK_PARAM, Callback.LINK_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        task_ids = await callback_helper.get_multi_params(args, Callback.TASK_ID_ARG, pre_command)

        if group_id is None or task_ids is None:
            return

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        tasks: list[Task] = []
        for task_id in task_ids:
            task = Task.get_from_id(task_id)
            if task is None:
                await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
                return
            tasks.append(task)

        message = "linked tasks"
        for task in tasks:
            group.link_task(task)
            message += " {}".format(task.get_id())

        message += " to TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_task_unlink(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.TASK_PARAM, Callback.UNLINK_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        task_ids = await callback_helper.get_multi_params(args, Callback.TASK_ID_ARG, pre_command)

        if group_id is None or task_ids is None:
            return
        
        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        tasks: list[Task] = []
        for task_id in task_ids:
            task = Task.get_from_id(task_id)
            if task is None:
                await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
                return

            tasks.append(task)

        message = "unlinked tasks"
        for task in tasks:
            message += " {}".format(task.get_id())
            group.unlink_task(task)

        message += " from TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_subtaskgroup(self, no_params: list[str], args: dict):
        if not no_params:
            await send_html("<code>{} {}</code> needs <code>{}</code>, <code>{}</code>, <code>{}</code> or <code>{}</code>".format(
                TaskGroupCallback.ACTION_NAME, Callback.SUBTASKGROUP_PARAM, Callback.CREATE_PARAM, Callback.GET_PARAM, Callback.LINK_PARAM, Callback.UNLINK_PARAM
            ))
            return

        action = no_params.pop(0)

        if action == Callback.CREATE_PARAM:
            await self.handle_subtaskgroup_create(no_params, args)
        elif action == Callback.GET_PARAM:
            await self.handle_subtaskgroup_get(no_params, args)
        elif action == Callback.LINK_PARAM:
            await self.handle_subtaskgroup_link(no_params, args)
        elif action == Callback.UNLINK_PARAM:
            await self.handle_subtaskgroup_unlink(no_params, args)
        else:
            await send_message("unknown param {}".format(action))

    async def handle_subtaskgroup_create(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.SUBTASKGROUP_PARAM, Callback.CREATE_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        subgroup_names = await callback_helper.get_multi_params(args, Callback.NAME_ARG, pre_command)

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        message = "created subgroups"
        for subgroup_name in subgroup_names:
            subgroup = group.add_subTaskGroup(subgroup_name)
            message += " {}".format(subgroup.get_id())
        
        message += " and added them to TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_subtaskgroup_get(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.SUBTASKGROUP_PARAM, Callback.GET_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        if group_id is None:
            return

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return
        
        message = """<h2>SubGroups of TaskGroup "{}"</h2>
<table>
    <tr>
        <th>SubGroup ID</th>
        <th>Name</th>
        <th>Has Tasks</th>
        <th>Has SubGroup</th>
    </tr>
        """.format(group.get_name())

        for subgroup in group.get_subTaskGroups():
            message += "<tr><td>{}</td><td>{}</td><td>{}</td><td>{}</td></tr>".format(
                subgroup.get_id(), subgroup.get_name(), bool(subgroup.get_tasks()), bool(subgroup.get_subTaskGroups()))

        message += "</table>"
        await send_html(message)

    async def handle_subtaskgroup_link(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.SUBTASKGROUP_PARAM, Callback.LINK_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        subgroup_ids = await callback_helper.get_multi_params(args, Callback.SUBTASKGROUP_ID_ARG, pre_command)

        if group_id is None or subgroup_ids is None:
            return

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        subgroups: list[TaskGroup] = []
        for subgroup_id in subgroup_ids:
            subgroup = TaskGroup.get_from_id(subgroup_id)
            if subgroup is None:
                await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
                return

            subgroups.append(subgroup)

        message = "linked subgroups"
        for subgroup in subgroups:
            group.link_subTaskGroup(subgroup)
            message += " {}".format(subgroup.get_id())

        message += " to TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_subtaskgroup_unlink(self, no_params: list[str], args: dict):
        pre_command = "{} {} {}".format(TaskGroupCallback.ACTION_NAME, Callback.SUBTASKGROUP_PARAM, Callback.UNLINK_PARAM)
        group_id = await callback_helper.get_single_param(args, Callback.TASKGROUP_ID_ARG, pre_command)
        subgroup_ids = await callback_helper.get_multi_params(args, Callback.SUBTASKGROUP_ID_ARG, pre_command)

        if group_id is None or subgroup_ids is None:
            return

        group = TaskGroup.get_from_id(group_id)
        if group is None:
            await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
            return

        subgroups: list[TaskGroup] = []
        for subgroup_id in subgroup_ids:
            subgroup = TaskGroup.get_from_id(subgroup_id)
            if subgroup is None:
                await send_message(TaskGroupCallback.WRONG_ID_MESSAGE)
                return

            subgroups.append(subgroup)

        message = "unlinked subgroups"
        for subgroup in subgroups:
            message += " {}".format(subgroup.get_id())
            group.unlink_subTaskGroup(subgroup)

        message += " from TaskGroup {}".format(group.get_id())
        await send_message(message)

    async def handle_interactive_set(self, no_params: list[str], args: dict):
        callback_helper.set_interactive(True, TaskGroupCallback)

    def _get_taskgroup_from_id(self, taskgroup_ids: list[str]):
        taskgroups = []
        if taskgroup_ids is not None:
            for taskgroup_id in taskgroup_ids:
                taskgroup = TaskGroup.get_from_id(taskgroup_id)
                if taskgroup is None:
                    return None

                taskgroups.append(taskgroup)
                
        else:
            with TaskGroup.tasksGroups_lock:
                taskgroups = TaskGroup.taskGroups

        return taskgroups

    def get_subtasks_group_str(self, taskgroup: TaskGroup, with_tasks=False):
        subtaskgroups = taskgroup.get_subTaskGroups()
        tasks = taskgroup.get_tasks()
        message = "<li>{}: {}".format(taskgroup.get_name(), taskgroup.get_id())
        if with_tasks:
            for task in tasks:
                message += "<br/>{} {}, subtasks={}: {}".format(Task.task_state_to_html(task.get_state()), task.get_name(), bool(task.get_subtasks()), task.get_id())
        for subtaskgroup in subtaskgroups:
            message += "<ul>{}</ul>".format(self.get_subtasks_group_str(subtaskgroup, with_tasks))

        message += "</li>"
        return message

