from email import message
import os
from nio import AsyncClient, MatrixRoom, RoomMessageText, Event
from nio.responses import SyncResponse
from callback import callback_helper



NEXT_BATCH_FILE = os.environ["NEXT_BATCH_PATH"]
STORE_PATH = os.environ["STORE_PATH"]
LOG_PATH = os.environ["LOG_PATH"]

BASE_URL = os.environ["BASE_URL"]
HOMESERVER_URL = "https://{}".format(BASE_URL)
ROOM_ID = "{}:{}".format(os.environ["ROOM_ID"], BASE_URL)
# DEVICE_ID = "wrapper_{}".format(str(uuid.uuid4()))

USERNAME = os.environ["USERNAME"]
PASSWORD = os.environ["PASSWORD"]

MIMETYPE_PLAIN = "text/plain"
MIMETYPE_HTML = "text/html"

M_ROOM_PREFIX = "m.room"
MESSAGE_TYPE_MESSAGE = "{}.message".format(M_ROOM_PREFIX)
# MESSAGE_TYPE_NOTICE = "{}.notice".format(M_ROOM_PREFIX)

_client: AsyncClient = None

async def message_callback(room: MatrixRoom, event: RoomMessageText) -> None:
    if room.room_id != ROOM_ID or event.sender == _client.user_id: #ignore other rooms and my messages
        return

    await callback_helper.exec(callback_helper.unpack(event.body))

async def sync_response_callback(response: SyncResponse) -> None:
    if isinstance(response, SyncResponse):
        await set_next_batch(response.next_batch)

async def default_callback(room: MatrixRoom, event: Event): # not sure if needed
    log("unhandled {} {} in room {}", event.__class__, event.event_id, room.room_id)

async def init():
    global _client
    _client = AsyncClient(HOMESERVER_URL, USERNAME, store_path=STORE_PATH)
    if not os.path.exists(STORE_PATH):
        os.mkdir(STORE_PATH)

async def send_message(content: str, mimetype: str=MIMETYPE_PLAIN):
    await _room_send(content, mimetype=mimetype)

async def send_html(content: str, sender=send_message):
    await sender(content, MIMETYPE_HTML)

async def _room_send(content: str, room_id: str=ROOM_ID, mimetype: str=MIMETYPE_PLAIN):
    global _client

    await _client.room_send(room_id, MESSAGE_TYPE_MESSAGE, {"msgtype": "m.text", "body": "", "mimetype": mimetype, "format": "org.matrix.custom.html", "formatted_body": content}, ignore_unverified_devices=True)

async def listen_messages():
    global _client
    next_batch = await get_next_batch() # must be called before callback registering

    _client.add_event_callback(message_callback, RoomMessageText)
    _client.add_response_callback(sync_response_callback)

    while True:
        try:
            await _client.sync_forever(timeout=1000, since=next_batch)
        except Exception as e:
            log(str(e))

async def set_next_batch(next_batch: str):
    with open(NEXT_BATCH_FILE, "w+") as f:
        f.write(next_batch)

async def get_next_batch():
    global _client
    next_batch = ""

    if os.path.exists(NEXT_BATCH_FILE):
        with open(NEXT_BATCH_FILE, "r+") as f:
            next_batch = f.readline()

    if next_batch == "":
        sync_response = await _client.sync(full_state=True)
        return sync_response.next_batch

    return next_batch

async def login():
    global _client
    await _client.login(PASSWORD)
    await _client.sync(full_state=True)
    
async def logout():
    global _client
    await _client.logout()

def log(content: str, *args):
    content = content.format(*args)
    with open("log.txt", "a+") as f:
        f.write("{}\n".format(content))

